// Date utility functions
import { CALENDAR_GREGORY } from '../constants/date';
import { RX_DATE, RX_DATE_SPLIT } from '../constants/regex';
import { concat } from './array';
import { identity } from './identity';
import { isDate, isString } from './inspect';
import { toInteger } from './number';

// --- Date utility methods ---

// Create or clone a date (`new Date(...)` shortcut)
export var createDate = function createDate() {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }
  return new Date(...args);
};

// Parse a date sting, or Date object, into a Date object (with no time information)
export var parseYMD = date => {
  if (isString(date) && RX_DATE.test(date.trim())) {
    var [year, month, day] = date.split(RX_DATE_SPLIT).map(v => toInteger(v, 1));
    return createDate(year, month - 1, day);
  } else if (isDate(date)) {
    return createDate(date.getFullYear(), date.getMonth(), date.getDate());
  }
  return null;
};

// Format a date object as `YYYY-MM-DD` format
export var formatYMD = date => {
  date = parseYMD(date);
  if (!date) {
    return null;
  }
  var year = date.getFullYear();
  var month = "0".concat(date.getMonth() + 1).slice(-2);
  var day = "0".concat(date.getDate()).slice(-2);
  return "".concat(year, "-").concat(month, "-").concat(day);
};

// Given a locale (or locales), resolve the browser available locale
export var resolveLocale = function resolveLocale(locales) /* istanbul ignore next */{
  var calendar = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : CALENDAR_GREGORY;
  locales = concat(locales).filter(identity);
  var fmt = new Intl.DateTimeFormat(locales, {
    calendar
  });
  return fmt.resolvedOptions().locale;
};

// Create a `Intl.DateTimeFormat` formatter function
export var createDateFormatter = (locale, options) => /* istanbul ignore next */{
  var dtf = new Intl.DateTimeFormat(locale, options);
  return dtf.format;
};

// Determine if two dates are the same date (ignoring time portion)
export var datesEqual = (date1, date2) => {
  // Returns true of the date portion of two date objects are equal
  // We don't compare the time portion
  return formatYMD(date1) === formatYMD(date2);
};

// --- Date "math" utility methods (for BCalendar component mainly) ---

export var firstDateOfMonth = date => {
  date = createDate(date);
  date.setDate(1);
  return date;
};
export var lastDateOfMonth = date => {
  date = createDate(date);
  date.setMonth(date.getMonth() + 1);
  date.setDate(0);
  return date;
};
export var addYears = (date, numberOfYears) => {
  date = createDate(date);
  var month = date.getMonth();
  date.setFullYear(date.getFullYear() + numberOfYears);
  // Handle Feb 29th for leap years
  if (date.getMonth() !== month) {
    date.setDate(0);
  }
  return date;
};
export var oneMonthAgo = date => {
  date = createDate(date);
  var month = date.getMonth();
  date.setMonth(month - 1);
  // Handle when days in month are different
  if (date.getMonth() === month) {
    date.setDate(0);
  }
  return date;
};
export var oneMonthAhead = date => {
  date = createDate(date);
  var month = date.getMonth();
  date.setMonth(month + 1);
  // Handle when days in month are different
  if (date.getMonth() === (month + 2) % 12) {
    date.setDate(0);
  }
  return date;
};
export var oneYearAgo = date => {
  return addYears(date, -1);
};
export var oneYearAhead = date => {
  return addYears(date, 1);
};
export var oneDecadeAgo = date => {
  return addYears(date, -10);
};
export var oneDecadeAhead = date => {
  return addYears(date, 10);
};

// Helper function to constrain a date between two values
// Always returns a `Date` object or `null` if no date passed
export var constrainDate = function constrainDate(date) {
  var min = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
  var max = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
  // Ensure values are `Date` objects (or `null`)
  date = parseYMD(date);
  min = parseYMD(min) || date;
  max = parseYMD(max) || date;
  // Return a new `Date` object (or `null`)
  return date ? date < min ? min : date > max ? max : date : null;
};